//===================== EXTENDED PAGE CONTROL ==================================
//
// For some reason Borland did not fully implement all of the tab styles in the
// TPageControl and TTabSheet components.  This extension fills in the remaining
// tab styles.  The most useful ones are the button style, the owner draw style
// and the ability to add glyphs onto each tab.
//
// The ownerdraw style is written to operate in exactly the same way as, say, the
// TListBox owner draw style.  So you can look up the help on that component's
// ownerdraw and apply the techniques to the TExPageControl. The TExPageControl
// has a canvas for each tab in the control to draw on.
//
// The ability to add glyphs to each tab is made by loading the glyphs into a
// TImageList and linking the list into the Glyphs property.  This technique
// is the native Windows API way of doing it, with messages set up to create
// the link. When an individual glyph is about to be used an event is called
// with the tab number and the glyph number, the glyph number can be altered by
// the event to allow a different glyph/tab mapping. An alternative to using a
// TImageList and the in built glyph methods would be to use the ownerdraw style.
//
// This component also illustrates how to make a property that is visible in the
// object inspector in the ancestor component, disappear in the derived component.
// A number of articles in the press have said you can't do this.  TPageControl
// has a MultiLine property.  In the TExPageControl, multiline is put into a
// styles set; to make the Multiline property disappear from the object inspector
// it is declared with no "write" statement definition.  It can still be accessed
// in the ancestor by using the "inherited" reserved word.
//
// Version 1.00
// Grahame Marsh 19 January 1997
//
// Freeware - you get it for free, I take nothing, I make no promises!
//
// Please feel free to contact me: grahame.s.marsh@corp.courtaulds.co.uk


unit
  ExPage;

interface

uses
  Windows, Messages, Classes, Controls, ComCtrls, StdCtrls, CommCtrl,
  SysUtils, DsgnIntf, LibConst;

// tab styles - search win32 api help for TCS_ for info on what each style does
type
  TTabStyle  = (tabButton, tabMultiline, tabRightJustify,
                tabIconLeft, tabLabelLeft, tabOwnerDraw);
  TTabStyles = set of TTabStyle;

// event to allow different mapping of glyphs from the imagelist component
type
  TGlyphMapEvent = procedure(Control: TWinControl; PageIndex : integer; var GlyphIndex : integer) of object;

type
  TExPageControl = class (TPageControl)
  private
    FCanvas : TControlCanvas;      // canvas for drawing on with tabOwnerDraw
    FGlyphs : TImageList;          // link to a TImageList component
    FTabStyles : TTabStyles;       // tab style
    FOnDrawItem : TDrawItemEvent;  // Owner draw event
    FOnGlyphMap : TGlyphMapEvent;  // glyph mapping event
    procedure SetGlyphs (Value : TImageList);
    procedure SetTabStyles (Value : TTabStyles);
    function GetMultiline : boolean;
    procedure CNDrawItem (var Msg : TWMDrawItem); message CN_DRAWITEM;
    procedure GlyphsChanged (Sender : TObject);
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure UpdateGlyphs; virtual;
    procedure CreateWnd; override;
    procedure DrawItem (Index: Integer; Rect: TRect; State: TOwnerDrawState); virtual;
    property MultiLine : boolean read GetMultiline; // republish Multiline as read only
  published
    constructor Create (AOwner : TComponent); override;
    destructor Destroy; override;
// canvas published    
    property Canvas : TControlCanvas read FCanvas write FCanvas;  // for owner draw
// link to TImageList
    property Glyphs : TImageList Read FGlyphs write SetGlyphs;
// tab styles property
    property Styles : TTabStyles read FTabStyles write SetTabStyles default [];
// owner draw event
    property OnDrawItem : TDrawItemEvent read FOnDrawItem write FOnDrawItem;
// glyph map event
    property OnGlyphMap : TGlyphMapEvent read FOnGlyphMap write FOnGlyphMap;
  end;

// redeclare TTabSheet so it can have a component editor declared here

type
  TExTabSheet = class (TTabSheet)
  end;

procedure Register;

//------------------------------------------------------------------------------

implementation

// If you programatically change tab styles it is possible to end up with a
// zero tab width (not good).  This is detected and this width is used.
const
  DefaultTabWidth = 100;

// glyph exception
procedure GlyphError;
begin
  raise EListError.Create ('TExPageControl error in setting tab glyph')
end;

// constructor must create a TControlCanvas for the owner draw style
constructor TExPageControl.Create (AOwner : TComponent);
begin
  inherited Create (AOwner);
  FCanvas := TControlCanvas.Create
end;

// remove link with glyphs and free the canvas
destructor TExPageControl.Destroy;
begin
  if Assigned (FGlyphs) then
    FGlyphs.OnChange := nil;
  FCanvas.Free;
  inherited Destroy
end;

// CreateParams called to set the additional style bits
procedure TExPageControl.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams (Params);

  with Params do
  begin
    if tabButton in FTabStyles then
      Style:= Style or TCS_BUTTONS;

    if tabRightJustify in FTabStyles then
      Style := Style or TCS_RIGHTJUSTIFY;

    if tabIconLeft in FTabStyles then
      Style := Style or TCS_FORCEICONLEFT;

    if tabLabelLeft in FTabStyles then
      Style := Style or TCS_FORCELABELLEFT;

    if tabOwnerDraw in FTabStyles then
      Style := Style or TCS_OWNERDRAWFIXED
  end
end;

// CreateWnd also must set links to the glyphs
procedure TExPageControl.CreateWnd;
begin
  inherited CreateWnd;
  if Assigned (FGlyphs) then
    SetGlyphs (FGlyphs)
end;

// if the glyphs should change then update the tabs
procedure TExPageControl.GlyphsChanged (Sender : TObject);
begin
  UpdateGlyphs
end;

// multiline property redefined as readonly, this makes it
// disappear from the object inspector
function TExPageControl.GetMultiline : boolean;
begin
  Result := inherited Multiline
end;

// link the tabs to the glyph list
// nil parameter removes link
procedure TExPageControl.SetGlyphs (Value : TImageList);
begin
  FGlyphs := Value;
  if Assigned(FGlyphs) then
  begin
    SendMessage (Handle, TCM_SETIMAGELIST, 0, FGlyphs.Handle);
    FGlyphs.OnChange := GlyphsChanged
  end else
    SendMessage (Handle, TCM_SETIMAGELIST, 0, 0);
  UpdateGlyphs
end;

// determine properties whenever the tab styles are changed
procedure TExPageControl.SetTabStyles (Value : TTabStyles);
begin
  if FTabStyles <> Value then
  begin
    FTabStyles := Value;

    if tabRightJustify in FTabStyles then
      FTabStyles := FTabStyles + [tabMultiLine];

    inherited Multiline := tabMultiline in FTabStyles;

    if tabLabelLeft in FTabStyles then
      FTabStyles := FTabStyles + [tabIconLeft];

    if (tabIconLeft in FTabStyles) and (TabWidth = 0) then
      TabWidth := DefaultTabWidth;

    ReCreateWnd
  end
end;

// update the glyphs linked to the tab
procedure TExPageControl.UpdateGlyphs;
var
  TCItem : TTCItem;
  Control,
  Loop : integer;
begin
  if FGlyphs <> nil then
  begin
    for Loop := 0 to pred(PageCount) do
    begin
      TCItem.Mask := TCIF_IMAGE;
      TCItem.iImage := Loop;
      Control := Loop;

// OnGlyphMap allows the user to reselect the glyph linked to a
// particular tab
      if Assigned (FOnGlyphMap) then
        FOnGlyphMap (Self, Control, TCItem.iImage);

      if SendMessage (Handle, TCM_SETITEM, Control, longint(@TCItem)) = 0 then
        GlyphError
    end
  end
end;

// called when Owner Draw style is selected:
// retrieve the component style, set up the canvas and
// call the DrawItem method
procedure TExPageControl.CNDrawItem (var Msg : TWMDrawItem);
var
  State: TOwnerDrawState;
begin
  with Msg.DrawItemStruct^ do
  begin
    State := TOwnerDrawState (WordRec (LongRec (itemState).Lo).Lo);
    FCanvas.Handle := hDC;
    FCanvas.Font := Font;
    FCanvas.Brush := Brush;
    if integer (itemID) >= 0 then
      DrawItem (itemID, rcItem, State)
    else
      FCanvas.FillRect (rcItem);
    FCanvas.Handle := 0
  end;
end;

// default DrawItem method
procedure TExPageControl.DrawItem (Index: Integer; Rect: TRect; State: TOwnerDrawState);
begin
  if Assigned (FOnDrawItem) then
    FOnDrawItem (Self, Index, Rect, State)
  else begin
    FCanvas.FillRect (Rect);
    if odSelected in State then
      FCanvas.TextOut (Rect.Left + 6, Rect.Top+4, Tabs[Index])
    else
      FCanvas.TextOut (Rect.Left + 2, Rect.Top, Tabs[Index])
  end
end;

//--- COMPONENT EDITOR ---------------------------------------------------------
// Provide a new PageControl component editor so that the glyphs are handled
// correctly.  Provides the same functions as the ancestor property editor:
// new, next and previous.  Enables glyphs to appear at design time.

type
  TExPageControlEditor = class(TDefaultEditor)
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

procedure TExPageControlEditor.ExecuteVerb(Index: Integer);
var
  PageControl: TExPageControl;
  Page: TExTabSheet;
  Designer: TFormDesigner;
begin
  if Component is TTabSheet then
    PageControl := TExTabSheet(Component).PageControl as TExPageControl
  else
    PageControl := TExPageControl(Component);

  if PageControl <> nil then
  begin
    Designer := Self.Designer;
    if Index = 0 then
    begin
      Page := TExTabSheet.Create(Designer.Form);
      try
        Page.Name := Designer.UniqueName(TExTabSheet.ClassName);
        Page.Parent := PageControl;
        Page.PageControl := PageControl
      except
        Page.Free;
        raise
      end;
      PageControl.ActivePage := Page;
      PageControl.UpdateGlyphs
    end else begin
      Page := PageControl.FindNextPage(PageControl.ActivePage,Index = 1, False) as TExTabSheet;
      if (Page <> nil) and (Page <> PageControl.ActivePage) then
        PageControl.ActivePage := Page
    end;

    Designer.SelectComponent(Page);
    Designer.Modified
  end
end;

function TExPageControlEditor.GetVerb(Index: Integer): string;
begin
  case Index of
    0 : Result := 'New (ex)page';
    1 : Result := 'Next page';
    2 : Result := 'Previous page'
  end
end;

function TExPageControlEditor.GetVerbCount: Integer;
begin
  Result := 3
end;

// Register the component and editor. I put the TExPageControl on the Win95
// palette along with TPageControl.  You can put it elsewhere if you like.
procedure Register;
begin
  RegisterComponents (LoadStr(srWin95), [TExPageControl]);

  RegisterClasses([TExTabSheet]);
  RegisterComponentEditor(TExPageControl, TExPageControlEditor);
  RegisterComponentEditor(TExTabSheet, TExPageControlEditor)
end;

end.


